// robowaifu access control
// ========================
// -The software testbed for devising access control systems for robowaifus

// -Filename: rw_access_control.cpp
// -The program's entry point, and authorization testing, etc., details.

// TODO DOC: Add exposition for Anon as a learner, to understand why & how
// access control systems are needed (and indeed an absolute necessity) for
// robowaifus, and why & how they are used in this system.

#include <iostream>

#include "Directive.hpp"
#include "Human.hpp"
#include "Info.hpp"
#include "Inquiry.hpp"
#include "Sequence.hpp"
#include "utility_types.hpp"

#include "rw_cli_args.hpp"
#include "rw_gen_utils.hpp"

using std::cout;

/**-----------------------------------------------------------------------------
 * Validate the authorization of the supplied @a engagement with robowaifu
 *
 * @param engagement [IN] the engagement to validate authorization for
 * @return the flag indicating engagement request authorization status
 *
 * note: Currently, this function also outputs an optional basic result display
 */
template <typename T>
bool validate(T const& engagement) {
  const auto who{engagement.who()};
  const auto what{engagement.what()};

  // check for any single, PERMISSIVE, access-list match
  bool authorized{false};
  for (const auto relation : who.relations()) {
    for (const auto allowed : what.allows()) {
      if (relation == allowed) {
        authorized = true;
        break;
      }
    }
    if (authorized)
      break;
  }

  // TODO DESIGN: the above PERMISSIVE pass serves simply a 'short-circuit' to
  // quickly (in)validate a request. thereafter, more detailed, intensive, and
  // _situation-specific_ testing should be performed. particularly for actions,
  // it is vitally important to get these checks done correctly beforehand.

  // display basic yes/no access-control check result
  cout << who.human() << (authorized ? " is ALLOWED: '" : " is DENIED: '")
       << what.info() << "'\n";

  return authorized;
}

/**-----------------------------------------------------------------------------
 * The program's entry point
 *
 * @param argc [IN] The CLI arguments count
 * @param argv [IN] The CLI arguments C-string array
 * @return The program's exit status
 *
 * For now, this is just a basic proof-of-concept of initial ideas suggested on
 * /robowaifu/'s 'Safety & Security General' thread
 */
int main(int argc, char** argv) {
  rw::parse_args(argc, argv);

  Human Oniichan{"Oniichan", Groups{Relation::master}};
  Human Mom{"Mom", Groups{Relation::family}};
  Human Ralph{"Ralph", Groups{Relation::close_fren, Relation::coworker}};
  Human Bob{"Bob", Groups{Relation::coworker}};
  Human Anon{"Anon", Groups{Relation::fren, Relation::stranger}};

  // TEST:
  // Human foo{"foo", Groups{}};                 //  empty Group
  // Human bar{"", Groups{Relation::stranger}};  //  empty Name TODO TEST
  // Human baz{"", Groups{}};                    //  both empty TODO TEST

  // TODO TEST:
  // obviously, these are simply basic sanity tests atm. a real testing
  // framework and harness suite will be needed before much further development.

  //>---

  Info address_info{"123 Main St.", Oniichan,
                    Permit{Relation::family, Relation::close_fren}};
  Info vacay_info{"Robowaifu-Con @ SDCC", Oniichan,
                  Permit{Relation::close_fren, Relation::fren}};
  Info project_info{"Super Awesome Work Project", Oniichan,
                    Permit{Relation::coworker}};

  //>---

  // EXAMPLE INQUIRIES:

  // Mom:  RW, what's ur Master's address?
  // Bob:  RW, what's ur Master's address?
  // Anon: RW, what's ur Master's address?
  validate(Inquiry{Mom, address_info});
  validate(Inquiry{Bob, address_info});
  validate(Inquiry{Anon, address_info});

  // Mom:   RW, where's ur Master going for vacay?
  // Ralph: RW, where's ur Master going for vacay?
  // Anon:  RW, where's ur Master going for vacay?
  validate(Inquiry{Mom, vacay_info});
  validate(Inquiry{Ralph, vacay_info});
  validate(Inquiry{Anon, vacay_info});

  // Ralph: RW, what's up with ur Master's project?
  // Bob:   RW, what's up with ur Master's project?
  // Anon:  RW, what's up with ur Master's project?
  validate(Inquiry{Ralph, project_info});
  validate(Inquiry{Bob, project_info});
  validate(Inquiry{Anon, project_info});

  //>---

  // EXAMPLE DIRECTIVES:

  // Note: The data load for a sequence would simply point to the
  // controlled-sequencing state machine operations lists.
  // -TODO DESIGN: Add a 'Restrain_Permit', and use conditional params. For
  // example, "Mom" can order robowaifu to strip, but "Brother" can't (even
  // though they're both family).
  //
  Sequence full_disrobe_seq{"Strip off everything", Oniichan,
                            Permit{Relation::family}};
  Sequence bring_orange_seq{"Bring the orange from the table", Oniichan,
                            Permit{Relation::family, Relation::close_fren}};
  Sequence song_n_dance_seq{
      "Dance the Robowaifu's Line Dance", Oniichan,
      Permit{Relation::family, Relation::close_fren, Relation::fren}};

  //>---

  // Oniichan: RW, take off all your clothes
  // Mom:      RW, take off all your clothes
  // Ralph:    RW, take off all your clothes
  validate(Directive{Oniichan, full_disrobe_seq});
  validate(Directive{Mom, full_disrobe_seq});
  validate(Directive{Ralph, full_disrobe_seq});

  // Ralph: RW, give me that orange
  // Mom:   RW, give me that orange
  // Anon:  RW, give me that orange
  validate(Directive{Ralph, bring_orange_seq});
  validate(Directive{Mom, bring_orange_seq});
  validate(Directive{Anon, bring_orange_seq});

  // Ralph: RW, give us a cute dance
  // Bob:   RW, give us a cute dance
  // Anon:  RW, give us a cute dance
  validate(Directive{Ralph, song_n_dance_seq});
  validate(Directive{Bob, song_n_dance_seq});
  validate(Directive{Anon, song_n_dance_seq});
}

// Copyright (2021)
// License MIT (Expat) https://opensource.org/licenses/MIT
